/****************************************************************************
*
*  mo640.cpp
*
*  Utility to convert FAT-Superfloppies with more than 512 bytes per sector
*  for usage with N512DASD.FLT
*
*  Release: 0.9
*  Update:  30.03.99
*  (C)      Marcel Mller 1999
*
****************************************************************************/


#define INCL_BASE
#define INCL_DOSDEVIOCTL
#include <os2.h>

#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <ctype.h>
#include <stdlib.h>


static int ssdest = 9;
static int rserial = 0;
static char* dev = NULL;


int sscheck(int ss)
{  switch(ss)
   {case 256:
      return 8;
    case 512:
      return 9;
    case 1024:
      return 10;
    case 2048:
      return 11;
    default:
      return -1;  // error
}  }

int main(int argc, char** argv)
{  if (argc < 2)
   {  puts("MO640 V 0.9\n"
           "(C) Marcel Mller 1999\n\n"
           "usage: MO640 <driveno>: [options]\n\n"
           "options: /s#  change logical sector size to #, default 512");
      return EXIT_FAILURE;
   }

   // parse command line
   while (--argc)
   {  char* cp = argv[argc];
      switch (cp[0])
      {case '/':
       case '-':
         switch (toupper(cp[1]))
         {case 'N':
            rserial = 1;
            continue;
          case 'S':
            ssdest = 0;
            sscanf(cp+2, "%u", &ssdest);
            ssdest = sscheck(ssdest);
            if (ssdest == -1)
            {  fprintf(stderr, "Sectorsize %u not supported.\n", ssdest);
               return 10;
            }
            continue;
          default:
            fprintf(stderr, "Unrecognized option %s.\n", cp);
            return 11;
         }
       default:
         if (dev)
         {  fprintf(stderr, "Command line syntax error at %s.\n", cp);
            return 11;
         }
         dev = cp;
   }  }

   HFILE f;
   ULONG l1,l2;
/*   APIRET rc = DosOpen(argv[1], &f, &l, 512, 0, OPEN_ACTION_CREATE_IF_NEW|OPEN_ACTION_OPEN_IF_EXISTS,
    OPEN_FLAGS_DASD|OPEN_FLAGS_NO_CACHE|OPEN_SHARE_DENYNONE|OPEN_ACCESS_READWRITE, NULL);
   if (rc != NO_ERROR)
   {  fprintf(stderr, "Failed to open device %s for writing. - Code %i.\n", argv[1], rc);
      return 2;
   }
*/

   APIRET rc;
   USHORT dh;
   char buf[32] = "";
   unsigned short buf2[16];

   rc = DosPhysicalDisk(INFO_GETIOCTLHANDLE, &dh, 2, dev, strlen(dev)+1);
   if (rc != NO_ERROR)
   {  fprintf(stderr, "Failed to get handle for physical device %s. - Code %i.\n", argv[1], rc);
      return 9;
   }

   try
   {  // lock phys. device
      l1 = 1;
      l2 = 31;
      rc = DosDevIOCtl(dh, 9, 0, buf, 32, &l1, buf, 32, &l2);
      if (rc != NO_ERROR)
         fprintf(stderr, "Warning: Failed to lock physical device %s. - Code %i.\n", argv[1], rc);

      // get device layout
      l1 = 1;
      l2 = 16;
      rc = DosDevIOCtl(dh, 9, 0x63, buf, 32, &l1, buf2, 32, &l2);
      if (rc != NO_ERROR)
      {  fprintf(stderr, "Failed to get parameters for physical device %s. - Code %i.\n", argv[1], rc);
         throw 7;
      }
      printf("\nPhysical device %s\n"
       "  Cylinders:             %i\n"
       "  Heads:                 %i\n"
       "  Sectors per track:     %i\n"
      , argv[1], buf2[1], buf2[2], buf2[3]);

      // read 1st block
      TRACKLAYOUT t = {1, 0, 0, 0, 1, {1, 512}};
      #pragma pack(1)
      struct bootblock
      {  char           branch[3];
         char           oemid[8];
         unsigned short bps;
         unsigned char  spc;
         unsigned short res;
         unsigned char  nfats;
         unsigned short ndirs;
         unsigned short nsects;
         unsigned char  media;
         unsigned short spf;
         unsigned short spt;
         unsigned short nsides;
         unsigned long  nhid;
         unsigned long  lnsects;
         unsigned short physdrv;
         unsigned char  sig;
         unsigned long  serial;
         char           name[11];
         char           fsid[8];
         char           bootcode[448];
         unsigned short magic;
      } boot;
      l1 = sizeof t;
      l2 = 512;
      rc = DosDevIOCtl(dh, 9, 0x64, &t, l1, &l1, &boot, 512, &l2);
      if (rc != NO_ERROR)
      {  fprintf(stderr, "Failed to read sector 0. - Code %i.\n", rc);
         throw 6;
      }

      // checkpoint
      if (boot.magic != 0xaa55)
      {  fprintf(stderr, "Error: illagal magic code %x. - Sector 0 seems to contain no valid data.\n", boot.magic);
         throw 5;
      }
      if (strncmp(boot.fsid, "FAT", 3))
      {  fprintf(stderr, "Error: illegal filesystem-ID %.8s. - This tool is only designed for FAT superfloppy volumes.\n", boot.fsid);
         throw 4;
      }

      // display parameters
      printf("\nBoot block information:\n"
       "  File system ID:        %.8s\n"
       "  OEM ID:                %.8s\n"
       "  Bytes per sector:      %i\n"
       "  Sectors per cluster:   %i\n"
       "  Reserved sectors:      %i\n"
       "  Number of FATs:        %i\n"
       "  Size of root directory:%i\n"
       "  Total sectors (short): %i\n"
       "  Media byte:            %X\n"
       "  Sectors per FAT:       %i\n"
       "  Sectors per track:     %i\n"
       "  Heads:                 %i\n"
       "  Hidden sectors:        %lu\n"
       "  Total sectors (long):  %lu\n"
       "  Serial number:         %lX\n"
       "  Volume name:           %.11s\n"
      , boot.fsid, boot.oemid, boot.bps, boot.spc, boot.res, boot.nfats, boot.ndirs, boot.nsects, boot.media
      , boot.spf, boot.spt, boot.nsides, boot.nhid, boot.lnsects, boot.serial, boot.name);

      // checkpoint 2
      int shift = sscheck(boot.bps);
      if (shift == -1)
      {  fprintf(stderr, "Sector size %u cannot be converted.\n", boot.bps);
         throw 3;
      }
      // log2 of conversion ratio
      shift -= ssdest;
      if (shift == 0)
      {  printf("\nDevice has %u bytes per sector. No changes made.\n", boot.bps);
         throw 0;
      }

      // direction
      if (shift > 0)
      {  // reduce sector size
         boot.bps >>= shift;
         boot.spc <<= shift;
         boot.res <<= shift;
         boot.spf <<= shift;
         boot.nsects <<= shift;
         boot.nhid <<= shift;
         boot.lnsects <<= shift;

      } else
      {  // increase sector size
         shift = -shift;
         int mask = (1 << shift) -1;
         // check
         if ( boot.spc & mask || boot.res & mask || boot.spf & mask
             || boot.nsects & mask || boot.nhid & mask || boot.lnsects & mask )
         {  fprintf(stderr, "Conversion impossible, because not all entries are multiples of %u.\n", mask+1);
            throw 2;
         }
         boot.bps <<= shift;
         boot.spc >>= shift;
         boot.res >>= shift;
         boot.spf >>= shift;
         boot.nsects >>= shift;
         boot.nhid >>= shift;
         boot.lnsects >>= shift;
      }

      // serial number
      if (rserial)
         boot.serial = rand() + (rand() << 11) + (rand() << 22);

      // save
      l1 = sizeof t;
      l2 = 512;
      rc = DosDevIOCtl(dh, 9, 0x44, &t, l1, &l1, &boot, 512, &l2);
      if (rc != NO_ERROR)
      {  fprintf(stderr, "Failed to write sector 0. - Code %i.\n", rc);
         throw 1;
      }

      // message
      printf("\nParameters changed:\n"
       "  Bytes per sector:      %i\n"
       "  Sectors per cluster:   %i\n"
       "  Reserved sectors:      %i\n"
       "  Sectors per FAT:       %i\n"
       "  Hidden sectors:        %lu\n"
       "  Total sectors (short): %i\n"
       "  Total sectors (long):  %lu\n"
       "  Serial number:         %lX\n"
       "\nDevice successfully converted.\n"
      , boot.bps, boot.spc, boot.res, boot.spf, boot.nhid, boot.nsects, boot.lnsects, boot.serial);

      // redetermine
      /**buf = 0;
      l1 = 1;
      l2 = 31;
      rc = DosDevIOCtl(dh, 8, 0x02, buf, 32, &l1, buf, 32, &l2);
      if (rc != NO_ERROR)
         fprintf(stderr, "Warning: Failed to redetermine media. - Code %i.\n", rc);*/

      throw EXIT_SUCCESS;
   } catch (int r)
   {  // unlock phys. device
      l1 = 1;
      l2 = 31;
      rc = DosDevIOCtl(dh, 9, 1, buf, 32, &l1, buf, 32, &l2);
      if (rc != NO_ERROR)
         fprintf(stderr, "Warning: Failed to unlock physical device %s. - Code %i.\n", argv[1], rc);
      return r;
}  }
